package se.chalmers.cs.gf.parse.chart;

import java.util.*;
import java.io.PrintWriter;

/**
 *  A bottom-up chart parser.
 */
public class KilburyParser {

        private CFGrammar grammar;

        /**
         *  Maps categories to the rules starting with each category.
         */
        private Map<Category,List<NonTerminalRule>> rulesByFirstCat;

        public KilburyParser(CFGrammar grammar) {
                this.grammar = grammar;
                this.rulesByFirstCat = buildRulesByFirstCat(grammar);
        }

        private static Map<Category,List<NonTerminalRule>> buildRulesByFirstCat(CFGrammar grammar) {
                Map<Category,List<NonTerminalRule>> map = new HashMap<Category,List<NonTerminalRule>>();
                for (NonTerminalRule r : grammar.getNonTerminalRules()) {
                        assert r.getSize() > 0;
                        Category start = r.getRhs().get(0);
                        List<NonTerminalRule> rs = map.get(start);
                        if (rs == null) {
                                rs = new LinkedList<NonTerminalRule>();
                                map.put(start, rs);
                        }
                        rs.add(r);
                }
                return map;
        }

        /**
         *  Get a list of the rules whose rhs start with the given category.
         *  @return A List of rules
         */
        private List<NonTerminalRule> rulesStartingWith(Category cat) {
                List<NonTerminalRule> rs = rulesByFirstCat.get(cat);
                return rs == null ? Collections.<NonTerminalRule>emptyList() : rs;
        }

        public Chart parse(List<Token> input) {
                int n = input.size();
                Chart chart = new Chart(n);

                // process first node without adding any tokens
                handleNode(chart, 0);

                int k = 0;
                for (Token t : input) {
                        k++;
//                        System.out.println("Token " + k + "(" + n + ")...");
                        // add edges for token ending at k
                        scan(chart, t, k);
                        // create all edges ending at k
                        handleNode(chart, k);
                }
                assert k == n : k + ", " + n;

//                 System.err.println("Grammar:");
//                 System.err.println(grammar);

//                 System.err.println("Active edges: ");
//                 System.err.println(chart.toString());

                chart.purgeActive();

                return chart;
        }

        /**
         *  Use the scan rule to add edges for the token between
         *  nodes k-1 and k.
         */
        private void scan(Chart chart, Token t, int k) {
                for (TerminalRule r : grammar.getTerminalRules()) {
                        if (r.matches(t)) {
                                TerminalEdge e = new TerminalEdge(r, t);
                                chart.add(e, k-1, k);
//                        System.out.println("(" + (k-1) + "," + k + "): " + e);
                        }
                }
        }

        /**
         *  Do all processing for a node for which the terminal ending at that
         *  node has been added.
         *  Adds edges for all empty rules and iterates predict and combine 
         *  to the least fixpoint.
         */
        private void handleNode(Chart chart, int k) {
                empty(chart, k);
                while (step(chart,k)) { }
        }

        /**
         *  Add edges for all empty rules at node k.
         */ 
        private void empty(Chart chart, int k) {
                // FIXME: we could keep a set of all the EmptyRules around and 
                // link to if from the chart
                for (EmptyRule r : grammar.getEmptyRules())
                        chart.add(new EmptyEdge(r), k, k);
        }

        /**
         *  Use the predict and combine rules once on all edges ending at k.
         *  Should be iterated to a fixed-point.
         *
         * @return true if a new edge was added, false otherwise.
         */ 
        private boolean step(Chart chart, int k) {
                boolean addedNew = false;
                for (int j = 0; j <= k; j++) {
//                        System.out.println("Checking " + j + " -> " + k);

                        // FIXME: hack to avoid concurrent modification in the predict rule
                        Set<Edge> s = new HashSet<Edge>(chart.get(j,k));
                        for (Edge e : s) {
                                if (!e.isActive()) {
                                        if (predict(chart, e, j, k))
                                                addedNew = true;
                                        if (combine(chart, e, j, k))
                                                addedNew = true;
                                }
                        }
                }
                return addedNew;
        }

        /**
         *  Use the combine rule on a passive edge from j to k.
         *
         *  @return true if a new edge was added, false otherwise.
         */ 
        private boolean combine(Chart chart, Edge e, int j, int k) {
                //               System.out.println("Combine...");

                boolean addedNew = false;
                for (int i = 0; i <= j; i++) { 
                        // FIXME: hack to avoid concurrent modification
                        Set<Edge> s = new HashSet<Edge>(chart.get(i,j));
                        for (Edge f : s) {
                                if (f.isActive() && f.needs(e.getCategory())) {
                                        NonTerminalEdge g = new NonTerminalEdge((NonTerminalEdge)f);
                                        g.stepForward();
                                        if (chart.add(g, i, k)) {
                                                addedNew = true;
//                                                System.out.println("(" + i + "," + k + "): " + g);
                                        }
                                }
                        }
                }
                return addedNew;
        }

        /**
         *  Use the predict rule on a passive edge from j to k.
         */
        private boolean predict(Chart chart, Edge e, int j, int k) {
//                System.out.println("Predict...");

                boolean addedNew = false;
                Category c = e.getCategory();
                for (NonTerminalRule r : rulesStartingWith(c)) {
//                       System.out.println("Predict (" + j + "," + k + "): " + r);

                        NonTerminalEdge f = new NonTerminalEdge(r);
                        f.stepForward();
                        if (chart.add(f, j, k)) {
                                addedNew = true;
//                                        System.out.println("(" + j + "," + k + "): " + f);
                        } 
//                       else System.err.println("DUPLICATE");
                }
                return addedNew;
        }

}
